function [flow] = coarse_to_fine(I1, I2, ws, warps, pyramid_levels, pyramid_factor)

[M N C] = size(I1);

width_Pyramid = cell(pyramid_levels,1);
height_Pyramid = cell(pyramid_levels,1);

I1_Pyramid = cell(pyramid_levels,C);
I2_Pyramid = cell(pyramid_levels,C);

% precompute image sizes
width_Pyramid{1} = N;
height_Pyramid{1} = M;

for i = 2:pyramid_levels
  width_Pyramid{i} = round(pyramid_factor*width_Pyramid{i-1});
  height_Pyramid{i} = round(pyramid_factor*height_Pyramid{i-1});
  if min(width_Pyramid{i}, height_Pyramid{i}) < 4
    pyramid_levels = i;
    break;
  end
end

% set up image pyramides
for i = 1:pyramid_levels
  if i == 1
    for j=1:C
      I1_Pyramid{1,j} = I1(:,:,j);
      I2_Pyramid{1,j} = I2(:,:,j);
    end
  else
    for j=1:C    
      I1_Pyramid{i,j} = imresize(I1_Pyramid{i-1,j}, [height_Pyramid{i} width_Pyramid{i}], 'bilinear');
      I2_Pyramid{i,j} = imresize(I2_Pyramid{i-1,j}, [height_Pyramid{i} width_Pyramid{i}], 'bilinear');
    end
  end
end

for level = pyramid_levels:-1:1;
  
  scale = pyramid_factor^(level-1);
  
  M = height_Pyramid{level};
  N = width_Pyramid{level};
 
  if level == pyramid_levels
 
    % TODO: better initialization
    u = ones(M,N)*0;
    v = ones(M,N)*0;
  else
    
    rescale_factor_u = width_Pyramid{level+1}/N;
    rescale_factor_v = height_Pyramid{level+1}/M;
    
    % remove outliers
    u = medfilt2(u, [3 3]);
    v = medfilt2(v, [3 3]);
  
    
    % prolongate to finer grid
    u = imresize(u,[M N], 'bilinear')/rescale_factor_u;    
    v = imresize(v,[M N], 'bilinear')/rescale_factor_v;
  end
  
  I1 = zeros(M,N,C);
  I2 = zeros(M,N,C);
  for j=1:C
    I1(:,:,j) = I1_Pyramid{level,j};
    I2(:,:,j) = I2_Pyramid{level,j};
  end
  
  fprintf('*** level = %d\n', level);
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  [u, v] = LK_flow(I1, I2, u, v, ws, warps);
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
  
end

% store final flow
flow = zeros(M,N,2);
flow(:,:,1) = u;
flow(:,:,2) = v;
